# -*- coding: utf-8 -*-

import bpy
import blf
import bmesh
import math
import itertools
from bpy_extras import view3d_utils


# 描画ユーティリティクラス
class PgToys_Draw:
    __handle = None

    # 2D描画コールバック登録
    @staticmethod
    def handle_add():
        PgToys_Draw.__handle = bpy.types.SpaceView3D.draw_handler_add(
            PgToys_Draw.draw_callback, (), "WINDOW", "POST_PIXEL"
        )

    # 2D描画コールバック削除
    @staticmethod
    def handle_remove():
        if PgToys_Draw.__handle is not None:
            bpy.types.SpaceView3D.draw_handler_remove(PgToys_Draw.__handle, "WINDOW")
            PgToys_Draw.__handle = None

    # 2D描画コールバック
    @staticmethod
    def draw_callback():
        # 編集モード以外なら終了
        if bpy.context.mode != "EDIT_MESH":
            return

        PgToys = bpy.context.scene.PgToysProperty
        region = bpy.context.region
        region_3d = bpy.context.space_data.region_3d

        # フォント設定
        ui_scale = bpy.context.preferences.system.ui_scale
        blf.size(0, PgToys.font_size * ui_scale)
        blf.color(0, *PgToys.font_color, 1.0)
        blf.shadow(0, 6, 0.0, 0.0, 0.0, 1.0)
        blf.enable(0, blf.SHADOW)

        # 頂点、辺、面選択モードか
        is_vert_mode, is_edge_mode, is_face_mode = bpy.context.tool_settings.mesh_select_mode[:]

        # 現在のモードのオブジェクトリスト
        sel_obj = bpy.context.objects_in_mode
        show_limit = PgToys.show_limit
        for obj in sel_obj:
            # メッシュ以外は無視
            if obj.type != "MESH":
                continue

            # 編集中のメッシュを取得
            bm = bmesh.from_edit_mesh(obj.data)
            matrix_world = obj.matrix_world

            # 頂点番号の表示
            if is_vert_mode and PgToys.show_vertex_index:
                # 選択中の頂点リスト
                selected_verts = [v for v in bm.verts if v.select]
                for v in selected_verts[:show_limit]:
                    PgToys_Draw._draw_vector_position(
                        region, region_3d, matrix_world @ v.co, str(v.index)
                    )
                show_limit = max(0, show_limit - len(selected_verts))
            # 辺番号の表示
            if is_edge_mode and PgToys.show_edge_index:
                # 選択中の辺リスト
                selected_edges = [e for e in bm.edges if e.select]
                for e in selected_edges[:show_limit]:
                    v0 = e.verts[0].co
                    v1 = e.verts[1].co
                    v = v0 + (v1 - v0) / 2
                    PgToys_Draw._draw_vector_position(
                        region, region_3d, matrix_world @ v, str(e.index)
                    )
                show_limit = max(0, show_limit - len(selected_edges))
            # 面番号の表示
            if is_face_mode and PgToys.show_face_index:
                # 選択中の面リスト
                selected_faces = [f for f in bm.faces if f.select]
                for f in selected_faces[:show_limit]:
                    v = f.calc_center_bounds()
                    PgToys_Draw._draw_vector_position(
                        region, region_3d, matrix_world @ v, str(f.index)
                    )
                show_limit = max(0, show_limit - len(selected_faces))
            # 辺の長さ表示
            if PgToys.show_edge_length:
                # 選択中の辺リスト
                if PgToys.show_linked:
                    selected_verts = [v for v in bm.verts if v.select]
                    selected_edges = {e for v in selected_verts for e in v.link_edges}
                else:
                    selected_edges = [e for e in bm.edges if e.select]
                for e in itertools.islice(selected_edges, show_limit):
                    v0 = e.verts[0].co
                    v1 = e.verts[1].co
                    delta = v1 - v0
                    v = v0 + delta / 2
                    l = delta.length
                    PgToys_Draw._draw_vector_position(
                        region,
                        region_3d,
                        matrix_world @ v,
                        PgToys_Draw._omit_float_value(l) + "m",
                    )
                show_limit = max(0, show_limit - len(selected_edges))
            # 角度表示
            if PgToys.show_angle:
                # 選択された頂点を取得(面に属しているもののみ)
                selected_verts = {v for v in bm.verts if v.select and v.link_faces}
                target_verts = selected_verts
                if PgToys.show_linked:
                    new_verts = set(selected_verts)
                    for vert in selected_verts:
                        for e in vert.link_edges:
                            new_verts.update(e.verts)
                    target_verts = new_verts

                # 対象の頂点がない
                if not selected_verts:
                    continue

                # 角度表示する面
                target_faces = {f for v in selected_verts if v.link_faces for f in v.link_faces}

                # 各選択頂点ごとに処理
                for vert in target_verts:
                    if show_limit == 0:
                        break

                    # エッジが2本未満
                    connected_edges = list(vert.link_edges)
                    if len(connected_edges) < 2:
                        continue

                    # 同じ面を共有するエッジのペアを探す
                    for e1, e2 in itertools.combinations(
                        connected_edges, 2
                    ):
                        if show_limit == 0:
                            break

                        # 2本のエッジの組み合わせを取得
                        v1, v2 = e1.other_vert(vert), e2.other_vert(vert)
                        if not v1 or not v2:
                            continue

                        # 共通の面があるかチェック
                        common_faces = (
                            # set(v1.link_faces) & set(v2.link_faces) & set(vert.link_faces)
                            set(v1.link_faces) & set(v2.link_faces) & set(vert.link_faces) & set(target_faces)
                        )
                        if not common_faces:
                            continue

                        # エッジの方向ベクトル取得
                        vec1 = (v1.co - vert.co).normalized()
                        vec2 = (v2.co - vert.co).normalized()

                        # 長さが0のベクトルをスキップ
                        if vec1.length == 0 or vec2.length == 0:
                            continue

                        # 角度計算
                        angle_rad = vec1.angle(vec2)
                        angle_deg = math.degrees(angle_rad)

                        offset = (vec1 + vec2).normalized() * 0.2
                        PgToys_Draw._draw_vector_position(
                            region,
                            region_3d,
                            matrix_world @ (vert.co + offset),
                            PgToys_Draw._omit_float_value(angle_deg) + "°",
                        )
                        show_limit = max(0, show_limit - 1)

        blf.disable(0, blf.SHADOW)

    # Vector位置に文字列描画
    @staticmethod
    def _draw_vector_position(region, region_3d, pos, text: str):
        # 2Dに投影変換
        co_2d = view3d_utils.location_3d_to_region_2d(region, region_3d, pos)
        # 番号描画
        if co_2d:
            w, h = blf.dimensions(0, text)
            blf.position(0, co_2d.x - w / 2, co_2d.y - h / 2, 0)
            blf.draw(0, text)

    # 整数の場合は小数点以下省略
    @staticmethod
    def _omit_float_value(value: float) -> str:
        formatted = f"{value:.2f}".rstrip("0").rstrip(".")
        return formatted if formatted != "" else "0"
